/*
 * kgraft_patch_bsc1012852
 *
 * Fix for CVE-2016-8632, bsc#1012852
 *
 *  Upstream commit:
 *  3de81b758853 ("tipc: check minimum bearer MTU")
 *
 *  SLE12-SP2 commit:
 *  b29a190a8a3dbc998daa5e8560b6c7b0670f892c
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/rtnetlink.h>
#include <linux/tipc_netlink.h>
#include <linux/inetdevice.h>
#include <linux/igmp.h>
#include <net/udp_tunnel.h>

#include "local.h"

#define KGR_PATCHED_MODULE      "tipc"

static int (*kgr_tipc_reset_bearer)(struct net *net, struct tipc_bearer *b_ptr);
static void (*kgr_bearer_disable)(struct net *net, struct tipc_bearer *b_ptr);
static const struct nla_policy (*kgr_tipc_nl_udp_policy)[TIPC_NLA_UDP_MAX + 1];
static void (*kgr_tipc_udp_media_addr_set)(struct tipc_media_addr *addr, struct udp_media_addr *ua);
static int (*kgr_tipc_udp_recv)(struct sock *sk, struct sk_buff *skb);
static int (*kgr_udp_sock_create4)(struct net *net, struct udp_port_cfg *cfg, struct socket **sockp);
static int (*kgr_udp_sock_create6)(struct net *net, struct udp_port_cfg *cfg, struct socket **sockp);
static void (*kgr_setup_udp_tunnel_sock)(struct net *net, struct socket *sock, struct udp_tunnel_sock_cfg *cfg);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "tipc_reset_bearer", (void *)&kgr_tipc_reset_bearer },
	{ "bearer_disable", (void *)&kgr_bearer_disable },
	{ "tipc_nl_udp_policy", (void *)&kgr_tipc_nl_udp_policy },
	{ "tipc_udp_media_addr_set", (void *)&kgr_tipc_udp_media_addr_set },
	{ "tipc_udp_recv", (void *)&kgr_tipc_udp_recv },
	{ "udp_sock_create4", (void *)&kgr_udp_sock_create4 },
	{ "udp_sock_create6", (void *)&kgr_udp_sock_create6 },
	{ "setup_udp_tunnel_sock", (void *)&kgr_setup_udp_tunnel_sock },
};

/* patched function */
int kgr_tipc_enable_l2_media(struct net *net, struct tipc_bearer *b,
	struct nlattr *attr[])
{
	struct net_device *dev;
	char *driver_name = strchr((const char *)b->name, ':') + 1;

	/* Find device with specified name */
	dev = dev_get_by_name(net, driver_name);
	if (!dev)
		return -ENODEV;
	/*
	 * Fix CVE-2016-8632
	 *  +4 lines
	 */
	if (kgr_tipc_check_mtu(dev, 0)) {
		dev_put(dev);
		return -EINVAL;
	}

	/* Associate TIPC bearer with L2 bearer */
	rcu_assign_pointer(b->media_ptr, dev);
	memset(&b->bcast_addr, 0, sizeof(b->bcast_addr));
	memcpy(b->bcast_addr.value, dev->broadcast, b->media->hwaddr_len);
	b->bcast_addr.media_id = b->media->type_id;
	b->bcast_addr.broadcast = 1;
	b->mtu = dev->mtu;
	b->media->raw2addr(b, &b->addr, (char *)dev->dev_addr);
	rcu_assign_pointer(dev->tipc_ptr, b);
	return 0;
}

/* patched function */
int kgr_tipc_l2_device_event(struct notifier_block *nb, unsigned long evt,
	void *ptr)
{
	struct net_device *dev = netdev_notifier_info_to_dev(ptr);
	struct net *net = dev_net(dev);
	struct tipc_bearer *b_ptr;

	b_ptr = rtnl_dereference(dev->tipc_ptr);
	if (!b_ptr)
		return NOTIFY_DONE;

	/*
	 * Fix CVE-2016-8632
	 *  -b_ptr->mtu = dev->mtu;
	 *  -
	 */
	switch (evt) {
	case NETDEV_CHANGE:
		if (netif_carrier_ok(dev))
			break;
	case NETDEV_GOING_DOWN:
	case NETDEV_CHANGEMTU:
		/*
		 * Fix CVE-2016-8632
		 *  +5 lines
		 */
		if (kgr_tipc_check_mtu(dev, 0)) {
			kgr_bearer_disable(net, b_ptr);
			break;
		}
		b_ptr->mtu = dev->mtu;
		kgr_tipc_reset_bearer(net, b_ptr);
		break;
	case NETDEV_CHANGEADDR:
		b_ptr->media->raw2addr(b_ptr, &b_ptr->addr,
				       (char *)dev->dev_addr);
		kgr_tipc_reset_bearer(net, b_ptr);
		break;
	case NETDEV_UNREGISTER:
	case NETDEV_CHANGENAME:
		kgr_bearer_disable(dev_net(dev), b_ptr);
		break;
	}
	return NOTIFY_OK;
}

/* inlined */
static int kgr_parse_options(struct nlattr *attrs[], struct udp_bearer *ub,
			 struct udp_media_addr *local,
			 struct udp_media_addr *remote)
{
	struct nlattr *opts[TIPC_NLA_UDP_MAX + 1];
	struct sockaddr_storage *sa_local, *sa_remote;

	if (!attrs[TIPC_NLA_BEARER_UDP_OPTS])
		goto err;
	if (nla_parse_nested(opts, TIPC_NLA_UDP_MAX,
			     attrs[TIPC_NLA_BEARER_UDP_OPTS],
			     *kgr_tipc_nl_udp_policy))
		goto err;
	if (opts[TIPC_NLA_UDP_LOCAL] && opts[TIPC_NLA_UDP_REMOTE]) {
		sa_local = nla_data(opts[TIPC_NLA_UDP_LOCAL]);
		sa_remote = nla_data(opts[TIPC_NLA_UDP_REMOTE]);
	} else {
err:
		pr_err("Invalid UDP bearer configuration");
		return -EINVAL;
	}
	if ((sa_local->ss_family & sa_remote->ss_family) == AF_INET) {
		struct sockaddr_in *ip4;

		ip4 = (struct sockaddr_in *)sa_local;
		local->proto = htons(ETH_P_IP);
		local->udp_port = ip4->sin_port;
		local->ipv4.s_addr = ip4->sin_addr.s_addr;

		ip4 = (struct sockaddr_in *)sa_remote;
		remote->proto = htons(ETH_P_IP);
		remote->udp_port = ip4->sin_port;
		remote->ipv4.s_addr = ip4->sin_addr.s_addr;
		return 0;

#if IS_ENABLED(CONFIG_IPV6)
	} else if ((sa_local->ss_family & sa_remote->ss_family) == AF_INET6) {
		struct sockaddr_in6 *ip6;

		ip6 = (struct sockaddr_in6 *)sa_local;
		local->proto = htons(ETH_P_IPV6);
		local->udp_port = ip6->sin6_port;
		local->ipv6 = ip6->sin6_addr;
		ub->ifindex = ip6->sin6_scope_id;

		ip6 = (struct sockaddr_in6 *)sa_remote;
		remote->proto = htons(ETH_P_IPV6);
		remote->udp_port = ip6->sin6_port;
		remote->ipv6 = ip6->sin6_addr;
		return 0;
#endif
	}
	return -EADDRNOTAVAIL;
}

/* inlined */
static int kgr_enable_mcast(struct udp_bearer *ub, struct udp_media_addr *remote)
{
	int err = 0;
	struct ip_mreqn mreqn;
	struct sock *sk = ub->ubsock->sk;

	if (ntohs(remote->proto) == ETH_P_IP) {
		if (!ipv4_is_multicast(remote->ipv4.s_addr))
			return 0;
		mreqn.imr_multiaddr = remote->ipv4;
		mreqn.imr_ifindex = ub->ifindex;
		err = ip_mc_join_group(sk, &mreqn);
#if IS_ENABLED(CONFIG_IPV6)
	} else {
		if (!ipv6_addr_is_multicast(&remote->ipv6))
			return 0;
		err = ipv6_stub->ipv6_sock_mc_join(sk, ub->ifindex,
						   &remote->ipv6);
#endif
	}
	return err;
}

/* inlined */
static inline int kgr_udp_sock_create(struct net *net,
				  struct udp_port_cfg *cfg,
				  struct socket **sockp)
{
	if (cfg->family == AF_INET)
		return kgr_udp_sock_create4(net, cfg, sockp);

	if (cfg->family == AF_INET6)
		return kgr_udp_sock_create6(net, cfg, sockp);

	return -EPFNOSUPPORT;
}

/* patched function */
int kgr_tipc_udp_enable(struct net *net, struct tipc_bearer *b,
	struct nlattr *attrs[])
{
	int err = -EINVAL;
	struct udp_bearer *ub;
	struct udp_media_addr *remote;
	struct udp_media_addr local = {0};
	struct udp_port_cfg udp_conf = {0};
	struct udp_tunnel_sock_cfg tuncfg = {NULL};

	ub = kzalloc(sizeof(*ub), GFP_ATOMIC);
	if (!ub)
		return -ENOMEM;

	remote = (struct udp_media_addr *)&b->bcast_addr.value;
	memset(remote, 0, sizeof(struct udp_media_addr));
	err = kgr_parse_options(attrs, ub, &local, remote);
	if (err)
		goto err;

	b->bcast_addr.media_id = TIPC_MEDIA_TYPE_UDP;
	b->bcast_addr.broadcast = 1;
	rcu_assign_pointer(b->media_ptr, ub);
	rcu_assign_pointer(ub->bearer, b);
	kgr_tipc_udp_media_addr_set(&b->addr, &local);
	if (local.proto == htons(ETH_P_IP)) {
		struct net_device *dev;

		dev = __ip_dev_find(net, local.ipv4.s_addr, false);
		if (!dev) {
			err = -ENODEV;
			goto err;
		}
		udp_conf.family = AF_INET;
		udp_conf.local_ip.s_addr = htonl(INADDR_ANY);
		udp_conf.use_udp_checksums = false;
		ub->ifindex = dev->ifindex;
		/*
		 * Fix CVE-2016-8632
		 *  +5 lines
		 */
		if (kgr_tipc_check_mtu(dev, sizeof(struct iphdr) +
					    sizeof(struct udphdr))) {
			err = -EINVAL;
			goto err;
		}
		b->mtu = dev->mtu - sizeof(struct iphdr)
			- sizeof(struct udphdr);
#if IS_ENABLED(CONFIG_IPV6)
	} else if (local.proto == htons(ETH_P_IPV6)) {
		udp_conf.family = AF_INET6;
		udp_conf.use_udp6_tx_checksums = true;
		udp_conf.use_udp6_rx_checksums = true;
		udp_conf.local_ip6 = in6addr_any;
		b->mtu = 1280;
#endif
	} else {
		err = -EAFNOSUPPORT;
		goto err;
	}
	udp_conf.local_udp_port = local.udp_port;
	err = kgr_udp_sock_create(net, &udp_conf, &ub->ubsock);
	if (err)
		goto err;
	tuncfg.sk_user_data = ub;
	tuncfg.encap_type = 1;
	tuncfg.encap_rcv = kgr_tipc_udp_recv;
	tuncfg.encap_destroy = NULL;
	kgr_setup_udp_tunnel_sock(net, ub->ubsock, &tuncfg);

	if (kgr_enable_mcast(ub, remote))
		goto err;
	return 0;
err:
	kfree(ub);
	return err;
}

static int kgr_patch_bsc1012852_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1012852_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1012852_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1012852_module_nb = {
	.notifier_call = kgr_patch_bsc1012852_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1012852_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1012852_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1012852_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1012852_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1012852_module_nb);
}
